// asm6502.c - copyright 1998-2004 Bruce Tomlin

//#define DEBUG_PASS
#define _65C02_

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

//#define OLD_OBJ			// use old object file format (instead of intel hex)
#define IHEX_SIZE	32		// max number of data bytes per line in intel hex format
#define maxSymLen	16		// max symbol length (only used in DumpSym())
#define maxOpcdLen	 5		// max opcode legnth (for building opcode table)
const int symTabCols = 3;	// number of columns for symbol table dump
#define maxMacParms	10		// maximum macro parameters

typedef unsigned char Str255[256];	// generic string type
typedef char bool;			// i guess bool is a c++ thing
const bool FALSE = 0;
const bool TRUE = 1;

const char		*progname;	// pointer to argv[0]
const Str255	version = "1.5";

enum instrType
{
	o_Illegal,	//  0 opcode not found in FindOpcode
	o_Implied,	//  1 implied instructions
	o_Branch,	//  2 branch instructions
	o_Mode,		//  3 instructions with multiple addressing modes
#ifdef _65C02_
	o_RSMB,		//    RMBn/SMBn instructions
	o_BBRS,		//    BBRn/BBSn instructions
#endif
	o_DB,		//  4 DB pseudo-op
	o_DW,		//  5 DW pseudo-op
	o_DS,		//  6 DS pseudo-op
	o_END,		//  7 END pseudo-op
	o_ENDM,		//  8 ENDM pseudo-op
	o_MacName,	//  9 Macro name
	o_FCC,		// 10 FCC pseudo-op
	o_LabelOp,	// 11 the following pseudo-ops handle the label field specially
	o_EQU,		// 12 EQU and SET pseudo-ops
	o_ORG,		// 13 ORG pseudo-op
	o_RORG,		// 14 RORG pseudo-op
	o_LIST,		// 15 LIST pseudo-op
	o_OPT,		// 16 OPT pseudo-op
	o_MACRO,	// 17 MACRO pseudo-op
};

enum addrMode
{
	a_None = -1, // -1 addressing mode invalid
	a_Imm,		//  0 Immediate    #val
	a_Abs,		//  1 Absolute     val
	a_Zpg,		//  2 Z-Page       val
	a_Acc,		//  3 Accumulator  A   (or no operand)
	a_Inx,		//  4 Indirect X   (val,X)
	a_Iny,		//  5 Indirect Y   (val),Y
	a_Zpx,		//  6 Z-Page X     val,X
	a_Abx,		//  7 Absolute X   val,X
	a_Aby,		//  8 Absolute Y   val,Y
	a_Ind,		//  9 Indirect     [val]
	a_Zpy,		// 10 Z-Page Y     val,Y
#ifdef _65C02_
	a_Zpi,		// 11 Z-Page Indirect  (val)
#endif
	a_Max		// 11/12
};

enum addrOps
{
	o_ORA,		//  0
	o_ASL,		//  1
	o_JSR,		//  2
	o_AND,		//  3
	o_BIT,		//  4
	o_ROL,		//  5
	o_EOR,		//  6
	o_LSR,		//  7
	o_JMP,		//  8
	o_ADC,		//  9
	o_ROR,		// 10
	o_STA,		// 11
	o_STY,		// 12
	o_STX,		// 13
	o_LDY,		// 14
	o_LDA,		// 15
	o_LDX,		// 16
	o_CPY,		// 17
	o_CMP,		// 18
	o_DEC,		// 19
	o_CPX,		// 20
	o_SBC,		// 21
	o_INC,		// 22
#ifdef _65C02_
	o_STZ,		// 23
	o_TSB,		// 24
	o_TRB,		// 25
#endif
	o_Max		// 23/26
};

#ifdef _65C02_
const unsigned char mode2op[] = // [o_Max * a_Max] =
{//Imm=0 Abs=1 Zpg=2 Acc=3 Inx=4 Iny=5 Zpx=6 Abx=7 Aby=8 Ind=9 Zpy=10 Zpi=11
	0x09, 0x0D, 0x05,    0, 0x01, 0x11, 0x15, 0x1D, 0x19,    0,    0, 0x12,	//  0 ORA
	   0, 0x0E, 0x06, 0x0A,    0,    0, 0x16, 0x1E,    0,    0,    0,    0,	//  1 ASL
	   0, 0x20,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,	//  2 JSR
	0x29, 0x2D, 0x25,    0, 0x21, 0x31, 0x35, 0x3D, 0x39,    0,    0, 0x32,	//  3 AND
	0x89, 0x2C, 0x24,    0,    0,    0, 0x34, 0x3C,    0,    0,    0,    0,	//  4 BIT
	   0, 0x2E, 0x26, 0x2A,    0,    0, 0x36, 0x3E,    0,    0,    0,    0,	//  5 ROL
	0x49, 0x4D, 0x45,    0, 0x41, 0x51, 0x55, 0x5D, 0x59,    0,    0, 0x52,	//  6 EOR
	   0, 0x4E, 0x46, 0x4A,    0,    0, 0x56, 0x5E,    0,    0,    0,    0,	//  7 LSR
	   0, 0x4C,    0,    0, 0x7C,    0,    0,    0,    0, 0x6C,    0,    0,	//  8 JMP note: 7C is (abs,x)
	0x69, 0x6D, 0x65,    0, 0x61, 0x71, 0x75, 0x7D, 0x79,    0,    0, 0x72,	//  9 ADC
	   0, 0x6E, 0x66, 0x6A,    0,    0, 0x76, 0x7E,    0,    0,    0,    0,	// 10 ROR
	   0, 0x8D, 0x85,    0, 0x81, 0x91, 0x95, 0x9D, 0x99,    0,    0, 0x92,	// 11 STA
	   0, 0x8C, 0x84,    0,    0,    0, 0x94,    0,    0,    0,    0,    0,	// 12 STY
	   0, 0x8E, 0x86,    0,    0,    0,    0,    0,    0,    0, 0x96,    0,	// 13 STX
	0xA0, 0xAC, 0xA4,    0,    0,    0, 0xB4, 0xBC,    0,    0,    0,    0,	// 14 LDY
	0xA9, 0xAD, 0xA5,    0, 0xA1, 0xB1, 0xB5, 0xBD, 0xB9,    0,    0, 0xB2,	// 15 LDA
	0xA2, 0xAE, 0xA6,    0,    0,    0,    0,    0, 0xBE,    0, 0xB6,    0,	// 16 LDX
	0xC0, 0xCC, 0xC4,    0,    0,    0,    0,    0,    0,    0,    0,    0,	// 17 CPY
	0xC9, 0xCD, 0xC5,    0, 0xC1, 0xD1, 0xD5, 0xDD, 0xD9,    0,    0, 0xD2,	// 18 CMP
	   0, 0xCE, 0xC6, 0x3A,    0,    0, 0xD6, 0xDE,    0,    0,    0,    0,	// 19 DEC
	0xE0, 0xEC, 0xE4,    0,    0,    0,    0,    0,    0,    0,    0,    0,	// 20 CPX
	0xE9, 0xED, 0xE5,    0, 0xE1, 0xF1, 0xF5, 0xFD, 0xF9,    0,    0, 0xF2,	// 21 SBC
	   0, 0xEE, 0xE6, 0x1A,    0,    0, 0xF6, 0xFE,    0,    0,    0,    0,	// 22 INC
	   0, 0x9C, 0x64,    0,    0,    0, 0x74, 0x9E,    0,    0,    0,    0,	// 23 STZ
	   0, 0x0C, 0x04,    0,    0,    0,    0,    0,    0,    0,    0,    0,	// 24 TSB
	   0, 0x1C, 0x14,    0,    0,    0,    0,    0,    0,    0,    0,    0	// 25 TRB
};
#else
const unsigned char mode2op[] = // [o_Max * a_Max] =
{//Imm=0 Abs=1 Zpg=2 Acc=3 Inx=4 Iny=5 Zpx=6 Abx=7 Aby=8 Ind=9 Zpy=10
	0x09, 0x0D, 0x05,    0, 0x01, 0x11, 0x15, 0x1D, 0x19,    0,    0,	//  0 ORA
	   0, 0x0E, 0x06, 0x0A,    0,    0, 0x16, 0x1E,    0,    0,    0,	//  1 ASL
	   0, 0x20,    0,    0,    0,    0,    0,    0,    0,    0,    0,	//  2 JSR
	0x29, 0x2D, 0x25,    0, 0x21, 0x31, 0x35, 0x3D, 0x39,    0,    0,	//  3 AND
	   0, 0x2C, 0x24,    0,    0,    0,    0,    0,    0,    0,    0,	//  4 BIT
	   0, 0x2E, 0x26, 0x2A,    0,    0, 0x36, 0x3E,    0,    0,    0,	//  5 ROL
	0x49, 0x4D, 0x45,    0, 0x41, 0x51, 0x55, 0x5D, 0x59,    0,    0,	//  6 EOR
	   0, 0x4E, 0x46, 0x4A,    0,    0, 0x56, 0x5E,    0,    0,    0,	//  7 LSR
	   0, 0x4C,    0,    0,    0,    0,    0,    0,    0, 0x6C,    0,	//  8 JMP
	0x69, 0x6D, 0x65,    0, 0x61, 0x71, 0x75, 0x7D, 0x79,    0,    0,	//  9 ADC
	   0, 0x6E, 0x66, 0x6A,    0,    0, 0x76, 0x7E,    0,    0,    0,	// 10 ROR
	   0, 0x8D, 0x85,    0, 0x81, 0x91, 0x95, 0x9D, 0x99,    0,    0,	// 11 STA
	   0, 0x8C, 0x84,    0,    0,    0, 0x94,    0,    0,    0,    0,	// 12 STY
	   0, 0x8E, 0x86,    0,    0,    0,    0,    0,    0,    0, 0x96,	// 13 STX
	0xA0, 0xAC, 0xA4,    0,    0,    0, 0xB4, 0xBC,    0,    0,    0,	// 14 LDY
	0xA9, 0xAD, 0xA5,    0, 0xA1, 0xB1, 0xB5, 0xBD, 0xB9,    0,    0,	// 15 LDA
	0xA2, 0xAE, 0xA6,    0,    0,    0,    0,    0, 0xBE,    0, 0xB6,	// 16 LDX
	0xC0, 0xCC, 0xC4,    0,    0,    0,    0,    0,    0,    0,    0,	// 17 CPY
	0xC9, 0xCD, 0xC5,    0, 0xC1, 0xD1, 0xD5, 0xDD, 0xD9,    0,    0,	// 18 CMP
	   0, 0xCE, 0xC6,    0,    0,    0, 0xD6, 0xDE,    0,    0,    0,	// 19 DEC
	0xE0, 0xEC, 0xE4,    0,    0,    0,    0,    0,    0,    0,    0,	// 20 CPX
	0xE9, 0xED, 0xE5,    0, 0xE1, 0xF1, 0xF5, 0xFD, 0xF9,    0,    0,	// 21 SBC
	   0, 0xEE, 0xE6,    0,    0,    0, 0xF6, 0xFE,    0,    0,    0	// 22 INC
};
#endif

struct SymRec
{
	struct SymRec	*next;		// pointer to next symtab entry
	unsigned short	value;		// symbol value
	bool			defined;	// TRUE if defined
	bool			multiDef;	// TRUE if multiply defined
	bool			isSet;		// TRUE if defined with SET pseudo
	bool			equ;		// TRUE if defined with EQU pseudo
	bool			known;		// TRUE if value is known
	char			name[1];	// symbol name, storage = 1 + length
} *symTab = NULL;		// pointer to first entry in symtab
typedef struct SymRec *SymPtr;

struct MacroLine
{
	struct MacroLine	*next;		// pointer to next macro line
	char				text[1];	// macro line, storage = 1 + length
};
typedef struct MacroLine *MacroLinePtr;

struct MacroParm
{
	struct MacroParm	*next;		// pointer to next macro parameter name
	char				name[1];	// macro parameter name, storage = 1 + length
};
typedef struct MacroParm *MacroParmPtr;

struct MacroRec
{
	struct MacroRec		*next;		// pointer to next macro
	bool				def;		// TRUE after macro is defined in pass 2
	MacroLinePtr		text;		// macro text
	MacroParmPtr		parms;		// macro parms
	int					nparms;		// number of macro parameters
	char				name[1];	// macro name, storage = 1 + length
} *macroTab = NULL;
typedef struct MacroRec *MacroPtr;

typedef char OpcdStr[maxOpcdLen+1];
struct OpcdRec
{
	OpcdStr			name;		// opcode name
	short			typ;		// opcode type
	unsigned short	parm;		// opcode parameter
} opcdTab[] =
{
	{"BRK",  o_Implied, 0x00},
	{"PHP",  o_Implied, 0x08},
	{"CLC",  o_Implied, 0x18},
	{"PLP",  o_Implied, 0x28},
	{"SEC",  o_Implied, 0x38},
	{"RTI",  o_Implied, 0x40},
	{"PHA",  o_Implied, 0x48},
	{"CLI",  o_Implied, 0x58},
	{"RTS",  o_Implied, 0x60},
	{"PLA",  o_Implied, 0x68},
	{"SEI",  o_Implied, 0x78},
	{"DEY",  o_Implied, 0x88},
	{"TXA",  o_Implied, 0x8A},
	{"TYA",  o_Implied, 0x98},
	{"TXS",  o_Implied, 0x9A},
	{"TAY",  o_Implied, 0xA8},
	{"TAX",  o_Implied, 0xAA},
	{"CLV",  o_Implied, 0xB8},
	{"TSX",  o_Implied, 0xBA},
	{"INY",  o_Implied, 0xC8},
	{"DEX",  o_Implied, 0xCA},
	{"CLD",  o_Implied, 0xD8},
	{"INX",  o_Implied, 0xE8},
	{"NOP",  o_Implied, 0xEA},
	{"SED",  o_Implied, 0xF8},

	{"ASLA", o_Implied, 0x0A},
	{"ROLA", o_Implied, 0x2A},
	{"LSRA", o_Implied, 0x4A},
	{"RORA", o_Implied, 0x6A},

	{"BPL",  o_Branch, 0x10},
	{"BMI",  o_Branch, 0x30},
	{"BVC",  o_Branch, 0x50},
	{"BVS",  o_Branch, 0x70},
	{"BCC",  o_Branch, 0x90},
	{"BCS",  o_Branch, 0xB0},
	{"BNE",  o_Branch, 0xD0},
	{"BEQ",  o_Branch, 0xF0},

	{"ORA",  o_Mode, o_ORA},
	{"ASL",  o_Mode, o_ASL},
	{"JSR",  o_Mode, o_JSR},
	{"AND",  o_Mode, o_AND},
	{"BIT",  o_Mode, o_BIT},
	{"ROL",  o_Mode, o_ROL},
	{"EOR",  o_Mode, o_EOR},
	{"LSR",  o_Mode, o_LSR},
	{"JMP",  o_Mode, o_JMP},
	{"ADC",  o_Mode, o_ADC},
	{"ROR",  o_Mode, o_ROR},
	{"STA",  o_Mode, o_STA},
	{"STY",  o_Mode, o_STY},
	{"STX",  o_Mode, o_STX},
	{"LDY",  o_Mode, o_LDY},
	{"LDA",  o_Mode, o_LDA},
	{"LDX",  o_Mode, o_LDX},
	{"CPY",  o_Mode, o_CPY},
	{"CMP",  o_Mode, o_CMP},
	{"DEC",  o_Mode, o_DEC},
	{"CPX",  o_Mode, o_CPX},
	{"SBC",  o_Mode, o_SBC},
	{"INC",  o_Mode, o_INC},

	{"DB",   o_DB,  0},
	{"FCB",  o_DB,  0},
	{"BYTE", o_DB,  0},
	{"DW",   o_DW,  0},
	{"FDB",  o_DW,  0},
	{"WORD", o_DW,  0},
	{"DS",   o_DS,  0},
	{"RMB",  o_DS,  0},
	{"BLKB", o_DS,  0},
	{"FCC",  o_FCC, 0},

	{"=",    o_EQU, 0},
	{"EQU",  o_EQU, 0},
	{"SET",  o_EQU, 1},

	{"ORG",  o_ORG,  0},
	{"AORG", o_ORG,  0},
	{"RORG", o_RORG, 0},
	{"END",  o_END,  0},
	{"LIST", o_LIST, 0},
	{"OPT",  o_OPT,  0},
	{"MACRO",o_MACRO,0},
	{"ENDM", o_ENDM, 0},

#ifdef _65C02_
	{"INCA", o_Implied, 0x1A},
	{"INA",  o_Implied, 0x1A},
	{"DECA", o_Implied, 0x3A},
	{"DEA",  o_Implied, 0x3A},
	{"PHY",  o_Implied, 0x5A},
	{"PLY",  o_Implied, 0x7A},
	{"PHX",  o_Implied, 0xDA},
	{"PLX",  o_Implied, 0xFA},

	{"BRA",  o_Branch, 0x80},

	{"STZ",  o_Mode, o_STZ},
	{"TSB",  o_Mode, o_TSB},
	{"TRB",  o_Mode, o_TRB},

	{"RMB0", o_RSMB, 0x07},
	{"RMB1", o_RSMB, 0x17},
	{"RMB2", o_RSMB, 0x27},
	{"RMB3", o_RSMB, 0x37},
	{"RMB4", o_RSMB, 0x47},
	{"RMB5", o_RSMB, 0x57},
	{"RMB6", o_RSMB, 0x67},
	{"RMB7", o_RSMB, 0x77},

	{"SMB0", o_RSMB, 0x87},
	{"SMB1", o_RSMB, 0x97},
	{"SMB2", o_RSMB, 0xA7},
	{"SMB3", o_RSMB, 0xB7},
	{"SMB4", o_RSMB, 0xC7},
	{"SMB5", o_RSMB, 0xD7},
	{"SMB6", o_RSMB, 0xE7},
	{"SMB7", o_RSMB, 0xF7},

	{"BBR0", o_BBRS, 0x0F},
	{"BBR1", o_BBRS, 0x1F},
	{"BBR2", o_BBRS, 0x2F},
	{"BBR3", o_BBRS, 0x3F},
	{"BBR4", o_BBRS, 0x4F},
	{"BBR5", o_BBRS, 0x5F},
	{"BBR6", o_BBRS, 0x6F},
	{"BBR7", o_BBRS, 0x7F},

	{"BBS0", o_BBRS, 0x8F},
	{"BBS1", o_BBRS, 0x9F},
	{"BBS2", o_BBRS, 0xAF},
	{"BBS3", o_BBRS, 0xBF},
	{"BBS4", o_BBRS, 0xCF},
	{"BBS5", o_BBRS, 0xDF},
	{"BBS6", o_BBRS, 0xEF},
	{"BBS7", o_BBRS, 0xFF},
#endif

	{"",     o_Illegal, 0}
};
typedef struct OpcdRec *OpcdPtr;

MacroPtr		macPtr;		// current macro in use
MacroLinePtr	macLine;	// current macro text pointer
char			*macParms[maxMacParms];	// pointers to current macro parameters
Str255			macParmsLine;	// text of current macro parameters

int		locPtr;			// Current program address
int		codPtr;			// Current program "real" address
int		pass;			// Current assembler pass
bool	warnFlag;		// TRUE if warning occurred this line
bool	errFlag;		// TRUE if error occurred this line
int		errCount;		// Total number of errors

Str255	line;			// Current line from input file
char	*linePtr;		// pointer into current line
Str255	listLine;		// Current listing line
bool	listFlag;		// FALSE to suppress listing source
bool	listThisLine;	// TRUE to force listing this line
bool	sourceEnd;		// TRUE when END pseudo encountered
Str255	lastLabl;		// last label for @ temps
bool	listMacFlag;	// FALSE to suppress showing macro expansions
bool	macLineFlag;	// TRUE if line came from a macro

unsigned char instr[5];		// Current instruction word
int		instrLen;		// Current instruction length

Str255	bytStr;			// Buffer for long DB statements
bool	showAddr;		// TRUE to show LocPtr on listing
int		xferAddr;		// Transfer address from END pseudo
bool	xferFound;		// TRUE if xfer addr defined w/ END

//	Command line parameters
Str255	cl_SrcName;		// Source file name
Str255	cl_ListName;	// Listing file name
Str255	cl_ObjName;		// Object file name
bool	cl_Err;			// TRUE for errors to screen
bool	cl_Warn;		// TRUE for warnings to screen
bool	cl_List;		// TRUE to generate listing file
bool	cl_Obj;			// TRUE to generate object file

FILE	*source;		// source input file
FILE	*object;		// object output file
FILE	*listing;		// listing output file

bool	evalKnown;	// TRUE if all operands in Eval were "known"


// --------------------------------------------------------------
// error messages


/*
 *	Error
 */

void Error(char *message)
{
	errFlag = TRUE;
	errCount++;

#ifndef DEBUG_PASS
	if (pass != 1)
#endif
	{
		listThisLine = TRUE;
		if (cl_List)	fprintf(listing,"*** Error:  %s ***\n",message);
		if (cl_Err)		fprintf(stdout/*stderr*/, "*** Error:  %s ***\n",message);
	}
}


/*
 *	Warning
 */

void Warning(char *message)
{
	if (pass != 1 && cl_Warn)
	{
		listThisLine = TRUE;
		warnFlag = TRUE;
		if (cl_List)	fprintf(listing,"*** Warning:  %s ***\n",message);
		if (cl_Warn)	fprintf(stdout/*stderr*/, "*** Warning:  %s ***\n",message);
	}
}


// --------------------------------------------------------------
// string utilities


/*
 *	Debleft deblanks the string s from the left side
 */

void Debleft(char *s)
{
	char *p = s;

	while (*p == 9 || *p == ' ')
		p++;

	while (*s++ = *p++);
}


/*
 *	Debright deblanks the string s from the right side
 */

void Debright(char *s)
{
	char *p = s + strlen(s);

	while (p>s && *--p == ' ')
		*p = 0;
}


/*
 *	Deblank removes blanks from both ends of the string s
 */

void Deblank(char *s)
{
	Debleft(s);
	Debright(s);
}


/*
 *	Uprcase converts string s to upper case
 */

void Uprcase(char *s)
{
	char *p = s;

	while (*p = toupper(*p))
		p++;
}


int ishex(char c)
{
	c = toupper(c);
	return isdigit(c) || ('A' <= c && c <= 'F');
}


int isalphanum(char c)
{
	c = toupper(c);
	return isdigit(c) || ('A' <= c && c <= 'Z') || c == '_';
}


int EvalBin(char *binStr)
{
	unsigned int	binVal;
	int				evalErr;
	int				c;

	evalErr = FALSE;
	binVal  = 0;

	while (c = *binStr++)
	{
		if (c < '0' || c > '1')
			evalErr = TRUE;
		else
			binVal = binVal * 2 + c - '0';
 	}

	if (evalErr)
	{
      binVal = 0;
      Error("Invalid binary number");
	}

   return binVal;
}


int EvalOct(char *octStr)
{
	unsigned int	octVal;
	int				evalErr;
	int				c;

	evalErr = FALSE;
	octVal  = 0;

	while (c = *octStr++)
	{
		if (c < '0' || c > '7')
			evalErr = TRUE;
		else
			octVal = octVal * 8 + c - '0';
 	}

	if (evalErr)
	{
      octVal = 0;
      Error("Invalid octal number");
	}

   return octVal;
}


int EvalDec(char *decStr)
{
	unsigned int	decVal;
	int				evalErr;
	int				c;

	evalErr = FALSE;
	decVal  = 0;

	while (c = *decStr++)
	{
		if (!isdigit(c))
			evalErr = TRUE;
		else
			decVal = decVal * 10 + c - '0';
 	}

	if (evalErr)
	{
      decVal = 0;
      Error("Invalid decimal number");
	}

   return decVal;
}


int EvalHex(char *hexStr)
{
	unsigned int	hexVal;
	int				evalErr;
	int				c;

	evalErr = FALSE;
	hexVal  = 0;

	while (c = toupper(*hexStr++))
	{
		if (!ishex(c))
			evalErr = TRUE;
		else
		{
			if (c > '9')	c = c - 'A' + 10;
			else			c = c - '0';
			hexVal = hexVal * 16 + c;
		}
 	}

	if (evalErr)
	{
      hexVal = 0;
      Error("Invalid hexadecimal number");
	}

	return hexVal;
}


// --------------------------------------------------------------
// token handling

// returns 0 for end-of-line, -1 for alpha-numeric, else char value for non-alphanumeric
int GetWord(char *word)
{
	unsigned char	c;

//	Deblank(line);
	word[0] = 0;

	// skip initial whitespace
	c = *linePtr;
	while (c == 12 || c == '\t' || c == ' ')
		c = *++linePtr;

	// skip comments
	if (c == ';')
		while (c)
			c = *++linePtr;

	// test for end of line
	if (c)
	{
		// test for alphanumeric token
		if (isalphanum(c))
		{
			while (isalphanum(c))
			{
				*word++ = toupper(c);
				c = *++linePtr;
			}
			*word = 0;
			return -1;
		}
		else
		{
			word[0] = c;
			word[1] = 0;
			linePtr++;
			return c;
		}
	}

	return 0;
}


void Expect(char *expected)
{
	Str255 s;
	GetWord(s);
	if (strcmp(s,expected) != 0)
	{
		sprintf(s,"\"%s\" expected\n",expected);
		Error(s);
	}
}


void Comma()
{
	Expect(",");
}


void RParen()
{
	Expect(")");
}


void EatIt()
{
	Str255	word;
	while (GetWord(word));		// eat junk at end of line
}


/*
 *	IllegalOperand
 */

void IllegalOperand()
{
	Error("Illegal operand");
	EatIt();
}


// --------------------------------------------------------------
// macro handling


MacroPtr FindMacro(char *name)
{
	MacroPtr p = macroTab;
	bool found = FALSE;

	while (p && !found)
	{
		found = (strcmp(p -> name, name) == 0);
		if (!found)
			p = p -> next;
	}

	return p;
}


MacroPtr AddMacro(char *name)
{
	MacroPtr	p;

	p = malloc(sizeof(struct MacroRec) + strlen(name));

	strcpy(p -> name, name);
	p -> def = FALSE;
	p -> text = NULL;
	p -> next = macroTab;
	p -> parms = NULL;
	p -> nparms = 0;

	macroTab = p;

	return p;
}
	

void AddMacroParm(MacroPtr macro, char *name)
{
	MacroParmPtr	parm;
	MacroParmPtr	p;

	parm = malloc(sizeof(struct MacroParm) + strlen(name));
	parm -> next = NULL;
	strcpy(parm -> name, name);
	macro -> nparms++;

	p = macro -> parms;
	if (p)
	{
		while (p -> next)
			p = p -> next;
		p -> next = parm;
	}
	else macro -> parms = parm;
}


void AddMacroLine(MacroPtr macro, char *line)
{
	MacroLinePtr	m;
	MacroLinePtr	p;

	m = malloc(sizeof(struct MacroLine) + strlen(line));
	m -> next = NULL;
	strcpy(m -> text, line);

	p = macro -> text;
	if (p)
	{
		while (p -> next)
			p = p -> next;
		p -> next = m;
	}
	else macro -> text = m;
}


void GetMacParms(MacroPtr macro)
{
	int		i;
	int		n;
	int		quote;
	char	c;
	char	*p;
	bool	done;

	for (i=0; i<maxMacParms; i++)
		macParms[i] = NULL;

	// skip initial whitespace
	c = *linePtr;
	while (c == 12 || c == '\t' || c == ' ')
		c = *++linePtr;

	// copy rest of line for safekeeping
	strcpy(macParmsLine, linePtr);

	n = 0;
	p = macParmsLine;
	while (*p && *p != ';' && n<maxMacParms)
	{
		// skip whitespace before current parameter
		c = *p;
		while (c == 12 || c == '\t' || c == ' ')
			c = *++p;

		// record start of parameter
		macParms[n] = p;
		n++;

		quote = 0;
		done = FALSE;
		// skip to end of parameter
		while(!done && (c = *p++))
		{
			switch(c)
			{
				case 0:
					--p;
					done = TRUE;
					break;

				case ';':
					if (quote==0)
					{
						*--p = 0;
						done = TRUE;
					}
					break;

				case ',':
					if (quote==0)
					{
						*(p-1) = 0;
						done = TRUE;
					}
					break;

				case 0x27:	// quote
				case '"':
					if (quote == 0)
						quote = c;
					else if (quote == c)
						quote = 0;
			}
		}
	}

	// terminate last parameter and point remaining parameters to null strings
	*p = 0;
	for (i=n; i<maxMacParms; i++)
		macParms[i] = p;

	// remove whitespace from end of parameter
	for (i=0; i<maxMacParms; i++)
		if (macParms[i])
		{
			p = macParms[i] + strlen(macParms[i]) - 1;
			while (p>=macParms[i] && (*p == ' ' || *p == 9))
				*p-- = 0;
		}

	if (n > macro -> nparms || n > maxMacParms)
		Error("Too many macro parameters");

//	for (i=0; i<maxMacParms; i++)
//		printf(" %d='%s'",i,macParms[i]);
//	printf("\n");
}


void DoMacParms()
{
	int				i;
//	int				j;
//	int				n;
	Str255			word;
	MacroParmPtr	parm;
	char			*p;		// pointer to start of word
	char			c;
	int				token;

//printf("  old line = '%s'\n",line);

	// start at beginning of line
	linePtr = line;

	// skip initial whitespace
	c = *linePtr;
	while (c == 12 || c == '\t' || c == ' ')
		c = *++linePtr;

	// while not end of line
	p = linePtr;
	token = GetWord(word);
	while (token)
	{
		// if alphanumeric, search for macro parameter of the same name
		if (token == -1)
		{
			i = 0;
			parm = macPtr -> parms;
			while (parm && strcmp(parm -> name, word))
			{
				parm = parm -> next;
				i++;
			}

			// if macro parameter found, replace parameter name with parameter value
			if (parm)
			{
//printf("found macro parameter name '%s', length=%d, index = %d\n",word, linePtr - p,i);
//printf("  old line = '%s'\n",line);
				// copy from linePtr to temp string
				strcpy(word, linePtr);
				// copy from corresponding parameter to p
				strcpy(p, macParms[i]);
				// point p to end of appended text
				p = p + strlen(macParms[i]);
				// copy from temp to p
				strcpy(p, word);
				// update linePtr
				linePtr = p;
//printf("  new line = '%s'\n",line);
//printf("  new linePtr = '%s'\n",linePtr);
			}
		}

		// skip initial whitespace
		c = *linePtr;
		while (c == 12 || c == '\t' || c == ' ')
			c = *++linePtr;

		p = linePtr;
		token = GetWord(word);
	}
//printf("  new line = '%s'\n",line);
}


void DumpMacro(MacroPtr p)
{
	MacroLinePtr	line;
	MacroParmPtr	parm;

	if (cl_List)
	{

	fprintf(listing,"--- Macro '%s' ---", p -> name);
	fprintf(listing," def = %d, nparms = %d\n", p -> def, p -> nparms);

//	dump parms here
	fprintf(listing,"Parms:");
	for (parm = p->parms; parm; parm = parm->next)
	{
		fprintf(listing," '%s'",parm->name);
	}
	fprintf(listing,"\n");

//	dump text here
	for (line = p->text; line; line = line->next)
			fprintf(listing," '%s'\n",line->text);
	}
}


void DumpMacroTab(void)
{
	struct	MacroRec *p;

	p = macroTab;
	while (p)
	{
		DumpMacro(p);
		p = p -> next;
	}
}


// --------------------------------------------------------------
// opcodes and symbol table


/*
 *	FindOpcode
 */

void FindOpcode(char *name, int *typ, int *parm, MacroPtr *macro)

{
	OpcdPtr p = opcdTab;
	bool found = FALSE;

	while (p -> typ != o_Illegal && !found)
	{
		found = (strcmp(p -> name, name) == 0);
		if (!found)
			p++;
	}

	*typ  = p -> typ;
	*parm = p -> parm;

	*macro = NULL;
	if (!found)
	{
		*macro = FindMacro(name);
		if (*macro)
			*typ = o_MacName;
	}
}


/*
 *	FindSym
 */

SymPtr FindSym(char *symName)
{
	SymPtr p = symTab;
	bool found = FALSE;

	while (p && !found)
	{
		found = (strcmp(p -> name, symName) == 0);
		if (!found)
			p = p -> next;
	}

	return p;
}


/*
 *	AddSym
 */

SymPtr AddSym(char *symName)
{
	SymPtr p;

	p = malloc(sizeof(struct SymRec) + strlen(symName));

	strcpy(p -> name, symName);
	p -> value = 0;
	p -> next = symTab;
	p -> defined = FALSE;
	p -> multiDef = FALSE;
	p -> isSet = FALSE;
	p -> equ = FALSE;
	p -> known = FALSE;

	symTab = p;

	return p;
}



/*
 *	RefSym
 */

int RefSym(char *symName, bool *known)
{
	SymPtr p;
	int i;
	Str255 s;

	if (p = FindSym(symName))
	{
		if (!p -> defined)
		{
			sprintf(s, "Symbol '%s' undefined", symName);
			Error(s);
		}
		switch(pass)
		{
			case 1:
				if (!p -> defined) *known = FALSE;
				break;
			case 2:
				if (!p -> known) *known = FALSE;
				break;
		}
		return p -> value;
	}

	{	// check for 'FFH' style constants here

		i = strlen(symName)-1;
		if (toupper(symName[i]) == 'H')
			while (i>0 && ishex(symName[i-1]))
				i--;

		if (i == 0)
		{
			strncpy(s, symName, 255);
			s[strlen(s)-1] = 0;
			return EvalHex(s);
		}
		else
		{
			p = AddSym(symName);
			*known = FALSE;
#ifdef DEBUG_PASS
			sprintf(s, "Symbol '%s' undefined", symName);
			Error(s);
#endif
		}
	}

	return 0;
}


/*
 *	DefSym
 */

void DefSym(char *symName, unsigned short val, bool setSym, bool equSym)
{
	SymPtr p;
	Str255 s;

	if (symName[0])	// if(strlen(symName) != 0)
	{
		p = FindSym(symName);
		if (p == NULL)
			p = AddSym(symName);

		if (!p -> defined || (p -> isSet && setSym))
		{
			p -> value = val;
			p -> defined = TRUE;
			p -> isSet = setSym;
			p -> equ = equSym;
		}
		else if (p -> value != val)
		{
			p -> multiDef = TRUE;
			if (pass == 2 && !p -> known)
				 sprintf(s, "Phase error");
			else sprintf(s, "Symbol '%s' multiply defined",symName);
			Error(s);
		}

		if (pass == 2) p -> known = TRUE;
	}
}


void DumpSym(SymPtr p, char *s)
{
	char *s2;
	int n;

	n = 0;

	s2 = p->name;
	while(*s2 && n<maxSymLen)
	{
		*s++ = *s2++;
		n++;
	}

	while (n < maxSymLen)
	{
		*s++ = ' ';
		n++;
	}

	sprintf(s, "%.4X ", p->value);

	s = s + strlen(s);

	n = 0;
	if (!p->defined)	{*s++ = 'U'; n++;}	// Undefined
	if ( p->multiDef)	{*s++ = 'M'; n++;}	// Multiply defined
	if ( p->isSet)		{*s++ = 'S'; n++;}	// Set
	if ( p->equ)		{*s++ = 'E'; n++;}	// Equ
	while (n < 9)
	{
		*s++ = ' ';
		n++;
	}
	*s = 0;
}


void DumpSymTab(void)
{
	struct	SymRec *p;
	int		i;
	Str255	s;

	i = 1;
	p = symTab;
	while (p)
	{
		DumpSym(p,s);
		p = p -> next;

		i++;
		if (p == NULL || i > symTabCols)
		{
			i = 1;
			Debright(s);
//			printf("%s\n",s);
			if (cl_List)
				fprintf(listing, "%s\n", s);
		}
		else
		{
//			printf("%s",s);
			if (cl_List)
				fprintf(listing, "%s", s);
		}
	}
}


void SortSymTab()
{
	SymPtr			i,j;	// pointers to current elements
	SymPtr			ip,jp;	// pointers to previous elements
	SymPtr			t;		// temp for swapping

	if (symTab != NULL)
	{
		ip = NULL;	i = symTab;
		jp = i;		j = i -> next;

		while (j != NULL)
		{
			while (j != NULL)
			{
				if (strcmp(i->name,j->name) > 0)	// (i->name > j->name)
				{
					if (ip != NULL) ip -> next = j;
							   else symTab     = j;
					if (i == jp)
					{
						i -> next = j -> next;
						j -> next = i;
					}
					else
					{
						jp -> next = i;

						t         = i -> next;
						i -> next = j -> next;
						j -> next = t;
					}
					t = i; i = j; j = t;
				}
				jp = j;		j = j -> next;
			}
			ip = i;		i = i -> next;
			jp = i;		j = i -> next;
		}
	}
}


// --------------------------------------------------------------
// expression evaluation


int Eval0(void);		// forward declaration


int Factor(void)
{
	Str255		word,s;
	int			token;
	int			val;

	token = GetWord(word);
	val = 0;

	switch(token)
	{
		case 0:
			Error("Missing operand");
			break;

		case '$':
			if (ishex(*linePtr))
			{
				GetWord(word);
				val = EvalHex(word);
				break;
			}
			// fall-through...
		case '*':
		case '.':
			val = locPtr;
			break;

		case '-':
			val = -Factor();
			break;

		case '+':
			val = Factor();
			break;

		case '~':
			val = ~Factor();
			break;

		case '<':
			val = Factor() & 0xFF;
			break;

		case '>':
			val = (Factor() >> 8) & 0xFF;
			break;

		case '%':
			GetWord(word);
			val = EvalBin(word);
			break;

		case '(':
			val = Eval0();
			RParen();
			break;

		case 0x27:	// single quote
			val = *linePtr;
			if (val)
			{
				linePtr++;
				Expect("'");
			}
			else Error("Missing operand");
           	break;

		case '@':
			GetWord(word);
			strcpy(s,lastLabl);
			strcat(s,"@");
			strcat(s,word);
			val = RefSym(s, &evalKnown);
			break;

		case -1:
			if (isdigit(word[0]))
			{
				val = strlen(word) - 1;
				token = word[val];
				switch(token)
				{
					case 'B':
						word[val] = 0;
						val = EvalBin(word);
						break;

					case 'O':
						word[val] = 0;
						val = EvalOct(word);
						break;

					case 'D':
						word[val] = 0;
						val = EvalDec(word);
						break;

					case 'H':
						word[val] = 0;
						val = EvalHex(word);
						break;

                    default:
						val = EvalDec(word);
						break;
              	}
			}
			else val = RefSym(word,&evalKnown);
			break;

		default:
			// ignore anything else
			break;
	}

  	return val;
}


int Term(void)
{
	Str255	word;
	int		token;
	int		val;
	char	*oldLine;

	val = Factor();

	oldLine = linePtr;
	token = GetWord(word);
	while (token == '*' || token == '/' || token == '%')
	{
		switch(token)
		{
			case '*':	val = val * Factor();	break;
			case '/':	val = val / Factor();	break;
			case '%':	val = val % Factor();	break;
		}
		oldLine = linePtr;
		token = GetWord(word);
	}
	linePtr = oldLine;

	return val;
}


int Eval1(void)
{
	Str255	word;
	int		token;
	int		val;
	char	*oldLine;

	val = Term();

	oldLine = linePtr;
	token = GetWord(word);
	while (token == '+' || token == '-')
	{
		switch(token)
		{
			case '+':	val = val + Term();		break;
			case '-':	val = val - Term();		break;
		}
		oldLine = linePtr;
		token = GetWord(word);
	}
	linePtr = oldLine;

	return val;
}


int Eval0(void)
{
	Str255	word;
	int		token;
	int		val;
	char	*oldLine;

	val = Eval1();

	oldLine = linePtr;
	token = GetWord(word);
	while (token == '&' || token == '|' ||
			(token == '<' && *linePtr == '<') ||
			(token == '>' && *linePtr == '>'))
	{
		switch(token)
		{
			case '&':	val = val & Eval1();	break;
			case '|':	val = val | Eval1();	break;
			case '<':	linePtr++;	val = val << Eval1();	break;
			case '>':	linePtr++;	val = val >> Eval1();	break;
		}
		oldLine = linePtr;
		token = GetWord(word);
	}
	linePtr = oldLine;

	return val;
}


int Eval(void)
{
	evalKnown = TRUE;

	return Eval0();
}


int EvalByte(void)
{
	int val;

	val = Eval();

	if (val < -128 || val > 255)
		Error("Byte out of range");

	return val & 255;
}


// --------------------------------------------------------------
// object file generation


#ifdef OLD_OBJ
int		newLoc;			// New program address
bool	updLoc;			// TRUE if newLoc needs to be written to file
#else
	unsigned char ihex_buf[IHEX_SIZE];
	int ihex_len;	// length of current intel hex line
	int ihex_base;	// current address in intel hex buffer
	int ihex_addr;	// start address of current intel hex line

// Intel hex format:
//
// :aabbbbccdddd...ddee
//
// aa    = record data length (the number of dd bytes)
// bbbb  = address for this record
// cc    = record type
//       00 = data (data is in the dd bytes)
//       01 = end of file (bbbb is transfer address)
//       02 = extended segment address record
//            dddd (two bytes) represents the segment address
//       03 = Start segment address record
//            dddd (two bytes) represents the segment of the transfer address
//       04 = extended linear address record
//            dddd (two bytes) represents the high address word
//       05 = Start linear address record
//            dddd (two bytes) represents the high word of the transfer address
// dd... = data bytes if record type needs it
// ee    = checksum byte: add all bytes aa through dd
//                        and subtract from 256 (2's complement negate)


void ihex_line(int addr, unsigned char *buf, int len, int rectype)
{
	int i,chksum;

	if (cl_Obj)
		{
		chksum = len + (addr >> 8) + addr + rectype;
		fprintf(object,":%.2X%.4X%.2X",len,addr & 0xFFFF,rectype);
		for (i=0; i<len; i++)
		{
			fprintf(object,"%.2X",buf[i]);
			chksum = chksum + buf[i];
		}
		fprintf(object,"%.2X\n",(-chksum) & 0xFF);
	}
}
#endif


void CodeInit(void)
{
#ifdef OLD_OBJ
	// Object file format does not need initialization
#else
	ihex_len  = 0;
	ihex_base = 0;
	ihex_addr = 0;
#endif
}


void CodeFlush(void)
{
#ifdef OLD_OBJ
	// Object file format does not use buffering; no flush needed
#else
	if (ihex_len)
	{
		ihex_line(ihex_base, ihex_buf, ihex_len, 0);
		ihex_len  = 0;
		ihex_base = ihex_base + ihex_len;
		ihex_addr = ihex_base;
	}
#endif
}


void CodeOut(int byte)
{
	if (pass == 2)
	{
#ifdef OLD_OBJ
		if (updLoc)
		{
			fprintf(object, ":%.4X\n",newLoc);	// codPtr? codPtr=newLoc?
			updLoc = FALSE;
		}

		fprintf(object, "%.2X\n", byte);
#else
		if (codPtr != ihex_addr)
		{
			CodeFlush();
			ihex_base = codPtr;
			ihex_addr = codPtr;
		}

		ihex_buf[ihex_len++] = byte;
		ihex_addr++;

		if (ihex_len == IHEX_SIZE)
			CodeFlush();
#endif
	}
	locPtr++;
	codPtr++;
}


void CodeEnd(void)
{
	CodeFlush();

	if (pass == 2)
	{
#ifdef OLD_OBJ
		if (xferFound)
			fprintf(object, "$%.4X\n",xferAddr);
#else
		if (xferFound)
			ihex_line(xferAddr, ihex_buf, 0, 1);
#endif
	}
}


void CodeAbsOrg(int addr)
{
	codPtr = addr;
	locPtr = addr;
#ifdef OLD_OBJ
	newLoc = codPtr;
	updLoc = TRUE;	// updLoc = (addr != oldCodPtr)
#endif
}


void CodeRelOrg(int addr)
{
	locPtr = addr;
}


void CodeXfer(int addr)
{
	xferAddr  = addr;
	xferFound = TRUE;
}


void AddLocPtr(int ofs)
{
#ifdef OLD_OBJ
	if (ofs) updLoc = TRUE;
#endif
	codPtr = codPtr + ofs;
	locPtr = locPtr + ofs;
}


void Instr1(int b)
{
	instr[0] = b;
	instrLen = 1;
}


void Instr2(int b1, int b2)
{
	instr[0] = b1;
	instr[1] = b2;
	instrLen = 2;
}


void Instr3(int b1, int b2, int b3)
{
	instr[0] = b1;
	instr[1] = b2;
	instr[2] = b3;
	instrLen = 3;
}


void Instr3W(int b, int w)
{
	Instr3(b,w & 255,w >> 8);
}


/*
void Instr4(int b1, int b2, int b3, int b4)
{
	instr[0] = b1;
	instr[1] = b2;
	instr[2] = b3;
	instr[3] = b4;
	instrLen = 4;
}


void Instr4W(int b1, int b2, int w)
{
	Instr4(b1,b2,w & 255,w >> 8);
}
*/


// --------------------------------------------------------------
// main assembler loops


void DoOpcode(int typ, int parm)
{
	int		val;
	int		i;
	Str255	word;
	char	*oldLine;
	int		token;
	char	ch;
	bool	done;
	int		opcode;
	int		mode;
	const unsigned char	*modes;		// pointer to current o_Mode instruction's opcodes

	switch(typ)
	{
		case o_Implied:
			Instr1(parm);
			break;

		case o_Branch:
			val = Eval();
			val = val - locPtr - 2;
            if (val < -128 || val > 127)
				Error("Branch out of range");
			Instr2(parm,val);
			break;

		case o_Mode:
			instrLen = 0;
			oldLine = linePtr;
			token = GetWord(word);

			modes = mode2op + parm * a_Max;

            mode = a_None;
            val  = 0;

			if (token == 0)
               mode = a_Acc;	// accumulator
			else
			{
				switch(token)
				{
					case '#':	// immediate
						val  = Eval();
						mode = a_Imm;
						break;

					case '[':	// indirect (this is a 6809ism that I got used to)
						val  = Eval();
						// should probably also handle 65C02 (abs,X) here
						Expect("]");
						mode = a_Ind;
						break;

					case '(':	// indirect X,Y
						val  = Eval();
						token = GetWord(word);
						switch (token)
						{
							case ',':	// (val,X)
								Expect("X");
								Expect(")");
								mode = a_Inx;
                             	break;

							case ')':	// (val) -and- (val),Y
								oldLine = line;
								token = GetWord(word);
								switch(token)
								{
									case 0:
										mode = a_Ind;
										if (val >= 0 && val < 256 && evalKnown &&
											(modes[a_Zpi] != 0))
													mode = a_Zpi;
											else	mode = a_Ind;
										break;
									case ',':
										Expect("Y");
										mode = a_Iny;
										break;
								}
                              	break;

							default:
								mode = a_None;
								break;
						}
						break;

					default:	// everything else
						if ((word[1] == 0) && (toupper(word[0]) == 'A'))
						{
							token = GetWord(word);
							if (token == 0)		// mustn't have anything after the 'A'
								mode = a_Acc;
						}
						else
						{
							linePtr = oldLine;
							val   = Eval();
							token = GetWord(word);

							switch(token)
							{
								case 0:		// abs or zpg
									if (val >= 0 && val < 256 && evalKnown &&
										(modes[a_Zpg] != 0))
												mode = a_Zpg;
										else	mode = a_Abs;
									break;

								case ',':
									token = GetWord(word);

									if ((word[1] == 0) && (toupper(word[0]) == 'X'))
									{
										if (val >= 0 && val < 256 &&
											(evalKnown || modes[a_Abx] == 0))
													mode = a_Zpx;
											else	mode = a_Abx;
									}
									else if ((word[1] == 0) && (toupper(word[0]) == 'Y'))
									{
										if (val >= 0 && val < 256 &&
											(evalKnown || modes[a_Aby] == 0)
												&& modes[a_Zpy] != 0)
													mode = a_Zpy;
											else	mode = a_Aby;
									}
							}
						}
				}
			}

			if (mode != a_None)
			{
				opcode = modes[mode];
				if (opcode == 0)
					mode = a_None;
			}

			instrLen = 0;
			switch(mode)
			{
				case a_None:  
					Error("Invalid addressing mode");
					break;

				case a_Acc:
					Instr1(opcode);
					break;

				case a_Inx:
					if (opcode == 0x7C)	// 65C02 JMP (abs,X)
					{
						Instr3W(opcode, val);
						break;
					}
					// else fall through

				case a_Imm:
				case a_Zpg:
				case a_Iny:
				case a_Zpx:
				case a_Zpy:
				case a_Zpi:
					Instr2(opcode, val & 0xFF);
					break;

				case a_Abs:
				case a_Abx:
				case a_Aby:
				case a_Ind:
					Instr3W(opcode, val);
					break;
			}
			break;

		case o_RSMB:		//    RMBn/SMBn instructions
			// RMB/SMB zpg
			val = Eval();
			Instr2(parm,val);
			break;

		case o_BBRS:		//    BBRn/BBSn instructions
			i = Eval();
			Expect(",");
			val = Eval();
			val = val - locPtr - 3;
            if (val < -128 || val > 127)
				Error("Branch out of range");
			Instr3(parm,i,val);
			break;

		case o_DB:
			instrLen = 0;
			oldLine = linePtr;
			token = GetWord(word);

			if (token == 0)
               Error("Missing operand");

			while (token)
			{
				if (token == '\'')
				{
					while (token == '\'')
					{	// FIXME: need to handle double quotes and backslashes here
						while (*linePtr != 0 && *linePtr != token)
							bytStr[instrLen++] = *linePtr++;
						token = *linePtr;
						if (token) linePtr++;
						if (token == '\'' && *linePtr == '\'')
							bytStr[instrLen++] = token;
						else
							token = *linePtr;
					 }
				}
				else
				{
					linePtr = oldLine;
					val = EvalByte();
					bytStr[instrLen++] = val;
				}

				token = GetWord(word);
				oldLine = linePtr;

				if (token == ',')
				{
					token = GetWord(word);
					if (token == 0)
						Error("Missing operand");
             	}
           	}
            instrLen = -instrLen;
			break;

		case o_DW:
            instrLen = 0;
            oldLine = linePtr;
            token = GetWord(word);

			if (token == 0)
               Error("Missing operand");

           	while (token)
			{
				linePtr = oldLine;
				val = Eval();
				bytStr[instrLen++] = val & 255;
				bytStr[instrLen++] = val >> 8;

				token = GetWord(word);
				oldLine = linePtr;

               	if (token == ',')
				{
					token = GetWord(word);
					if (token == 0)
						Error("Missing operand");
               	}
           	}
            instrLen = -instrLen;
			break;

		case o_DS:
			val = Eval();

			if (pass == 2)
			{
				showAddr = FALSE;

				sprintf(word,"%.4X: (%.4X)",locPtr,val);
				for (i=0; i<12; i++)
					listLine[i] = word[i];
			}

			AddLocPtr(val);
			break;

		case o_END:
			oldLine = linePtr;
			if (GetWord(word))
			{
				linePtr = oldLine;
				val = Eval();
				CodeXfer(val);

				sprintf(word,"----  (%.4X)",val);
				for (i=6; i<12; i++)
					listLine[i] = word[i];
			}
			sourceEnd = TRUE;
			break;

		case o_ENDM:
			Error("ENDM without MACRO");
			break;

		case o_FCC:
			instrLen = 0;
            oldLine = linePtr;
            token = GetWord(word);

			if (token == 0)
               Error("Missing operand");

			ch = token;

			done = FALSE;
			while (token && !done)
			{
				if (token == ch)
				{
					while (token == ch)
					{	while (*linePtr != 0 && *linePtr != token)
							bytStr[instrLen++] = *linePtr++;
						token = *linePtr;
						if (token) linePtr++;
						if (token == '\'' && *linePtr == '\'')
							bytStr[instrLen++] = token;
						else
							token = *linePtr;
					 }
				}
				else
				{
					linePtr = oldLine;
					val = EvalByte();
					bytStr[instrLen++] = val;
				}

				token = GetWord(word);
				oldLine = linePtr;

				if (token == ',')
				{
					token = GetWord(word);
					if (token == 0)
						Error("Missing operand");
             	}
           	}
            instrLen = -instrLen;

			if (!done)
				Error("FCC not terminated properly");

			break;
/*
            WHILE (Length(line)<>0) AND NOT done DO BEGIN
               IF line[1]=ch THEN BEGIN
                  IF (Length(line)>1) AND (line[2]=ch) THEN BEGIN
                     bytStr := bytStr + line[1];
                     Delete(line,1,2);
                  END
                  ELSE BEGIN
                     Delete(line,1,1);
                     DebLeft(line);

                     IF (Length(line)=0) OR (line[1]=comment)
                        THEN done := TRUE
                        ELSE line := '';
                  END;
               END
               ELSE BEGIN
                  bytStr := bytStr + line[1];
                  Delete(line,1,1);
               END;
            END;
*/
		default:
			Error("Unknown opcode");
			EatIt();
			break;
	}
}


int ReadLine(FILE *file, char *line, int max)
{
	int c;
	int len = 0;

	macLineFlag = (macLine != NULL);
	if (macLineFlag)
	{
		strcpy(line, macLine -> text);
		macLine = macLine -> next;
		DoMacParms();
	}
	else
	{
		macPtr = NULL;

		while (max > 1)
		{
			c = fgetc(file);
			*line = 0;
			switch(c)
			{
				case EOF:
					if (len == 0) return 0;
				case '\n':
					return 1;
				case '\r':
					break;
				default:
					*line++ = c;
					max--;
					len++;
					break;
			}
		}
		while (c != EOF && c != '\n')
			c = fgetc(file);
	}
	return 1;
}


void ListOut(void)
{
/*	FIXME: need to handle form feed properly
	IF Deblank(listLine) = #12 THEN
		WriteLn(listing,#12)
*/
	Debright(listLine);

	if (cl_List)
		fprintf(listing,"%s\n",listLine);

	if (listLine[0])
	{
		if ((errFlag && cl_Err) || warnFlag)
			printf("%s\n",listLine);
	}
}


void DoLabelOp(int typ, int parm, char *labl)
{
	int			val;
	int			i;
	Str255		word;
	int			token;
	Str255		opcode;
	MacroPtr	macro;
	MacroPtr	xmacro;
	int			nparms;

	switch(typ)
	{
		case o_EQU:
			if (labl[0] == 0)
				Error("Missing label");
			else
			{
				val = Eval();

				sprintf(word,"---- = %.4X",val);
				for (i=5; i<11; i++)
					listLine[i] = word[i];

				DefSym(labl,val,parm==1,parm==0);
			}
			break;

		case o_ORG:
			CodeAbsOrg(Eval());
			DefSym(labl,locPtr,FALSE,FALSE);
			showAddr = TRUE;
			break;

		case o_RORG:
			val = Eval();
			CodeRelOrg(val);
			DefSym(labl,/*locPtr?*/codPtr,FALSE,FALSE);

            if (pass == 2)
			{
				sprintf(word,"%.4X = %.4X",codPtr,val);
				for (i=0; i<11; i++)
					listLine[i] = word[i];
			}
			break;

		case o_LIST:
			listThisLine = TRUE;

			if (labl[0])
				Error("Label not allowed");

			GetWord(word);

			     if (strcmp(word,"ON") == 0)		listFlag = TRUE;
			else if (strcmp(word,"OFF") == 0)		listFlag = FALSE;
			else if (strcmp(word,"MACRO") == 0)		listMacFlag = TRUE;
			else if (strcmp(word,"NOMACRO") == 0)	listMacFlag = FALSE;
			else									IllegalOperand();

			break;

		case o_OPT:
			listThisLine = TRUE;

			if (labl[0])
				Error("Label not allowed");

			GetWord(word);

			     if (strcmp(word,"LIST") == 0)		listFlag = TRUE;
			else if (strcmp(word,"NOLIST") == 0)	listFlag = FALSE;
			else if (strcmp(word,"MACRO") == 0)		listMacFlag = TRUE;
			else if (strcmp(word,"NOMACRO") == 0)	listMacFlag = FALSE;
			else									Error("Illegal option");

			break;

		case o_MACRO:
			macro = FindMacro(labl);
			if (macro && macro -> def)
				Error("Macro multiply defined");
			else
			{
				if (macro == NULL)
				{
					macro = AddMacro(labl);
					nparms = 0;

					token = GetWord(word);
					while (token == -1)
					{
						nparms++;
						if (nparms > maxMacParms)
							Error("Too many macro parameters");
						else
						{
							AddMacroParm(macro,word);
							token = GetWord(word);
							if (token == ',')
								token = GetWord(word);
						}
					}


					if (word[0])
						Error("Illegal operand");
             	}

				if (pass == 2)
					macro -> def = TRUE;

				i = ReadLine(source, line, sizeof(line));
				while (i && typ != o_ENDM)
				{
					if (pass == 2)
						ListOut();
					strcpy(listLine, "                ");		// 16 blanks
					strncat(listLine, line, 255-16);

					// skip initial formfeeds
					linePtr = line;
					while (*linePtr == 12)
						linePtr++;

					labl[0] = 0;
					if (isalphanum(*linePtr) || *linePtr == '@')
					{
						token = GetWord(labl);
						if (token)
							showAddr = TRUE;
							while (*linePtr == ' ' || *linePtr == '\t')
								linePtr++;

						if (labl[0])
						{
							if (token == '@')
							{
								GetWord(word);
								strcpy(labl,lastLabl);
								strcat(labl,"@");
								strcat(labl,word);			// labl = lastLabl + "@" + word;
							}
							else
								strcpy(lastLabl,labl);
						}

						if (*linePtr == ':')
							linePtr++;
					}

					typ = 0;
					if (GetWord(opcode))
						FindOpcode(opcode, &typ, &parm, &xmacro);

					switch(typ)
					{
						case o_END:
							Error("END not allowed inside a macro");
							break;

						case o_ENDM:
							if (pass == 1 && labl[0])
								AddMacroLine(macro,labl);
							break;

						default:
							if (pass == 1)
								AddMacroLine(macro,line);
							break;
					}
					i = ReadLine(source, line, sizeof(line));
             	}

				if (typ != o_ENDM)
					Error("Missing ENDM");
			}
			break;

		default:
			Error("Unknown opcode");
			EatIt();
			break;
  	}
}


void DoPass()
{
	Str255		labl;
	Str255		opcode;
	int			typ;
	int			parm;
	int			i;
	Str255		word;
	int			token;
	MacroPtr	macro;

	fseek(source, 0, SEEK_SET);	// rewind source file
	sourceEnd = FALSE;
	lastLabl[0] = 0;

	printf("Pass %d\n",pass);
#ifdef DEBUG_PASS
	if (cl_List)
		fprintf(listing,"========== Pass %d ==========\n",pass);
#endif

	CodeAbsOrg(0);
	errCount = 0;
	listFlag = TRUE;
	listMacFlag = FALSE;
	i = ReadLine(source, line, sizeof(line));
	while (i && !sourceEnd)
	{
		errFlag      = FALSE;
		warnFlag     = FALSE;
		instrLen     = 0;
		showAddr     = FALSE;
		listThisLine = listFlag;
		strcpy(listLine, "                ");		// 16 blanks
		strncat(listLine, line, 255-16);

		// skip initial formfeeds
		linePtr = line;
		while (*linePtr == 12)
			linePtr++;

		// look for label at beginning of line
		labl[0] = 0;
		if (isalphanum(*linePtr) || *linePtr == '@')
		{
			token = GetWord(labl);
			if (token)
				showAddr = TRUE;
			while (*linePtr == ' ' || *linePtr == '\t')
				linePtr++;

            if (labl[0])
			{
				if (token == '@')
				{
					GetWord(word);
					strcpy(labl,lastLabl);
					strcat(labl,"@");
					strcat(labl,word);			// labl = lastLabl + "@" + word;
				}
				else
					strcpy(lastLabl,labl);
			}

			if (*linePtr == ':')
				linePtr++;
		}

		token = GetWord(opcode);
		if (token == 0 || token == '*')
		{
			DefSym(labl,locPtr,FALSE,FALSE);
		}
		else
		{
			FindOpcode(opcode, &typ, &parm, &macro);

			if (typ == o_Illegal)
			{
				sprintf(word,"Illegal opcode '%s'",opcode);
				Error(word);
			}
			else if (typ == o_MacName)
			{
				if (macPtr)
					Error("Nested macros not supported");
				else
				{
					macPtr  = macro;
					macLine = macro -> text;

					GetMacParms(macro);

					showAddr = TRUE;
					DefSym(labl,locPtr,FALSE,FALSE);
				}
			}
			else if (typ >= o_LabelOp)
			{
				showAddr = FALSE;
				DoLabelOp(typ,parm,labl);
			}
			else
			{
				showAddr = TRUE;
				DefSym(labl,locPtr,FALSE,FALSE);
				DoOpcode(typ, parm);
			}

			if (typ != o_Illegal && typ != o_MacName)
				if (GetWord(word))
					Error("Too many operands");
		}

#ifndef DEBUG_PASS
		if (pass == 1)
			AddLocPtr(abs(instrLen));
		else
#endif
		{
			if (showAddr)
			{
				sprintf(word,"%.4X",locPtr);	// codPtr
				for (i=0; i<4; i++)
					listLine[i] = word[i];
			}

			if (instrLen>0)
				for (i = 0; i < instrLen; i++)
				{
					sprintf(word,"%.2X",instr[i]);
					listLine[i*2+5] = word[0];
					listLine[i*2+6] = word[1];
					CodeOut(instr[i]);
			}
			else if (instrLen<0)
			{
				for (i = 0; i < -instrLen; i++)
				{
					if (i > 0 && i % 5 == 0)
					{
						if (listThisLine)
							ListOut();
						strcpy(listLine, "                ");		// 16 blanks
						sprintf(word,"%.4X",locPtr);
						listLine[0] = word[0];
						listLine[1] = word[1];
						listLine[2] = word[2];
						listLine[3] = word[3];
					}
					sprintf(word,"%.2X",bytStr[i]);
					if (i<5)
					{
						listLine[i*2+5] = word[0];
						listLine[i*2+6] = word[1];
					}
					else
					{
						listLine[(i%5)*2+5] = word[0];
						listLine[(i%5)*2+6] = word[1];
						listLine[(i%5)*2+7] = 0;
					}
					CodeOut(bytStr[i]);
				}
			}

			if (listThisLine && (errFlag || listMacFlag || !macLineFlag))
				ListOut();
		}

		i = ReadLine(source, line, sizeof(line));
	}

	if (pass == 2) CodeEnd();

	// Put the lines after the END statement into the listing file
	// while still checking for listing control statements.  Ignore
	// any lines which have invalid syntax, etc., because whatever
	// is found after an END statement should esentially be ignored.

	if (pass == 2)
	{
		while (i)
		{
			listThisLine = listFlag;
			strcpy(listLine, "                ");		// 16 blanks
			strncat(listLine, line, 255-16);

			if (line[0]==' ' || line[0]=='\t')			// ignore labels (this isn't the right way)
			{
				GetWord(word);
				FindOpcode(opcode, &typ, &parm, &macro);
				if (typ == o_LIST || typ == o_OPT)
				{
					DoLabelOp(typ,parm,"");
				}
			}

			if (listThisLine)
				ListOut();

			i = ReadLine(source, line, sizeof(line));
		}
	}

}

// --------------------------------------------------------------
// initialization and parameters


void usage(void)
{
	printf("6502 assembler version %s\n",version);
	printf("\n");
	printf("Usage:\n");
	printf("    %s [options] srcfile\n",progname);
	printf("\n");
	printf("Options:\n");
	printf("    --             end of options\n");
	printf("    -e             show errors to screen\n");
	printf("    -w             show warnings to screen\n");
	printf("    -l [filename]  make a listing file, default is srcfile.lst\n");
#ifdef OLD_OBJ
	printf("    -o [filename]  make an object file, default is srcfile.obj\n");
#else
	printf("    -o [filename]  make an object file, default is srcfile.hex\n");
#endif
	exit(1);
}


void getopts(int argc, char * const argv[])
{
	int ch;

	while ((ch = getopt(argc, argv, "ewl:o:?")) != -1)
	{
		switch (ch)
		{
			case 'e':
				cl_Err = TRUE;
				break;

			case 'w':
				cl_Warn = TRUE;
				break;

			case 'l':
				cl_List = TRUE;
				if (optarg[0] =='-')
				{
					optarg = "";
					optind--;
				}
				strncpy(cl_ListName, optarg, 255);
				break;

			case 'o':
				cl_Obj = TRUE;
				if (optarg[0] =='-')
				{
					optarg = "";
					optind--;
				}
				strncpy(cl_ObjName, optarg, 255);
				break;

			case '?':
			default:
				usage();
		}
	}
	argc -= optind;
	argv += optind;

	// now argc is the number of remaining arguments
	// and argv[0] is the first remaining argument

	if (argc != 1)
		usage();

	strncpy(cl_SrcName, argv[0], 255);

	if (cl_List && cl_ListName[0] == 0)
	{
		strncpy(cl_ListName, cl_SrcName, 255-4);
		strcat (cl_ListName, ".lst");
	}

	if (cl_Obj  && cl_ObjName [0] == 0)
	{
		strncpy(cl_ObjName, cl_SrcName, 255-4);
#ifdef OLD_OBJ
		strcat (cl_ObjName, ".obj");
#else
		strcat (cl_ObjName, ".hex");
#endif
	}
}


int main (int argc, char * const argv[])
{
	// initialize and get parms

	progname = argv[0];
	pass = 0;
	symTab = NULL;
	xferAddr = 0;
	xferFound = FALSE;

	macroTab  = NULL;
	macPtr    = NULL;
	macLine   = NULL;

	cl_Err  = FALSE;
	cl_Warn = FALSE;
	cl_List = FALSE;
	cl_Obj  = FALSE;

	cl_SrcName [0] = 0;		source  = NULL;
	cl_ListName[0] = 0;		listing = NULL;
	cl_ObjName [0] = 0;		object  = NULL;

	getopts(argc, argv);

//	printf("cl_Err = %d, cl_Warn = %d, cl_List = %d, cl_Obj = %d\n",
//			cl_Err, cl_Warn, cl_List, cl_Obj);
//	printf("cl_SrcName = '%s', cl_ListName = '%s', cl_ObjName = '%s'\n",
//			cl_SrcName, cl_ListName, cl_ObjName);

	// open files

	source = NULL;
	source = fopen(cl_SrcName, "r");
	if (source == NULL)
	{
		printf("Unable to open source input file '%s'!\n",cl_SrcName);
		exit(1);
	}

	listing = NULL;
	if (cl_List)
	{
		listing = fopen(cl_ListName, "w");
		if (listing == NULL)
		{
			printf("Unable to create listing output file '%s'!\n",cl_ListName);
			if (source)
				fclose(source);
			exit(1);
		}
	}

	object = NULL;
	if (cl_Obj)
	{
		object = fopen(cl_ObjName, "w");
		if (object == NULL)
		{
			printf("Unable to create object output file '%s'!\n",cl_ObjName);
			if (source)
				fclose(source);
			if (listing)
				fclose(listing);
			exit(1);
		}
	}

	CodeInit();

	pass = 1;
	DoPass();

#ifdef DEBUG_PASS
	SortSymTab();
	DumpSymTab();
#endif

	pass = 2;
	DoPass();

	if (cl_List)	fprintf(listing, "\n%.5d Total Error(s)\n\n", errCount);
	if (cl_Err)		printf("\n%.5d Total Error(s)\n\n", errCount);

	SortSymTab();
	DumpSymTab();
//	DumpMacroTab();

	if (source)
		fclose(source);
	if (listing)
		fclose(listing);
	if (object)
		fclose(object);

	return (errCount != 0);
}